# Copyright 2016 Free Software Foundation, Inc.
#
# This file is part of GNU Radio
#
# GNU Radio is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# GNU Radio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Radio; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.

from __future__ import absolute_import

import re
import shlex

# back port from python3

_find_unsafe = re.compile(r'[^\w@%+=:,./-]').search


def _shlex_quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


if not hasattr(shlex, 'quote'):
    quote = _shlex_quote
else:
    quote = shlex.quote

split = shlex.split



"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import

from os import path

from gi.repository import Gtk

from . import Constants, Utils, Dialogs


class FileDialogHelper(Gtk.FileChooserDialog, object):
    """
    A wrapper class for the gtk file chooser dialog.
    Implement a file chooser dialog with only necessary parameters.
    """
    title = ''
    action = Gtk.FileChooserAction.OPEN
    filter_label = ''
    filter_ext = ''

    def __init__(self, parent, current_file_path):
        """
        FileDialogHelper constructor.
        Create a save or open dialog with cancel and ok buttons.
        Use standard settings: no multiple selection, local files only, and the * filter.

        Args:
            action: Gtk.FileChooserAction.OPEN or Gtk.FileChooserAction.SAVE
            title: the title of the dialog (string)
        """
        ok_stock = {
            Gtk.FileChooserAction.OPEN: 'gtk-open',
            Gtk.FileChooserAction.SAVE: 'gtk-save'
        }[self.action]

        Gtk.FileChooserDialog.__init__(self, title=self.title, action=self.action,
                                       transient_for=parent)
        self.add_buttons('gtk-cancel', Gtk.ResponseType.CANCEL, ok_stock, Gtk.ResponseType.OK)
        self.set_select_multiple(False)
        self.set_local_only(True)

        self.parent = parent
        self.current_file_path = current_file_path or path.join(
            Constants.DEFAULT_FILE_PATH, Constants.NEW_FLOGRAPH_TITLE + Constants.FILE_EXTENSION)

        self.set_current_folder(path.dirname(current_file_path))  # current directory
        self.setup_filters()

    def setup_filters(self, filters=None):
        set_default = True
        filters = filters or ([(self.filter_label, self.filter_ext)] if self.filter_label else [])
        filters.append(('All Files', ''))
        for label, ext in filters:
            if not label:
                continue
            f = Gtk.FileFilter()
            f.set_name(label)
            f.add_pattern('*' + ext)
            self.add_filter(f)
            if not set_default:
                self.set_filter(f)
                set_default = True

    def run(self):
        """Get the filename and destroy the dialog."""
        response = Gtk.FileChooserDialog.run(self)
        filename = self.get_filename() if response == Gtk.ResponseType.OK else None
        self.destroy()
        return filename


class SaveFileDialog(FileDialogHelper):
    """A dialog box to save or open flow graph files. This is a base class, do not use."""
    action = Gtk.FileChooserAction.SAVE

    def __init__(self, parent, current_file_path):
        super(SaveFileDialog, self).__init__(parent, current_file_path)
        self.set_current_name(path.splitext(path.basename(self.current_file_path))[0] + self.filter_ext)
        self.set_create_folders(True)
        self.set_do_overwrite_confirmation(True)


class OpenFileDialog(FileDialogHelper):
    """A dialog box to save or open flow graph files. This is a base class, do not use."""
    action = Gtk.FileChooserAction.OPEN

    def show_missing_message(self, filename):
        Dialogs.MessageDialogWrapper(
            self.parent,
            Gtk.MessageType.WARNING, Gtk.ButtonsType.CLOSE, 'Cannot Open!',
            'File <b>{filename}</b> Does not Exist!'.format(filename=Utils.encode(filename)),
        ).run_and_destroy()

    def get_filename(self):
        """
        Run the dialog and get the filename.
        If this is a save dialog and the file name is missing the extension, append the file extension.
        If the file name with the extension already exists, show a overwrite dialog.
        If this is an open dialog, return a list of filenames.

        Returns:
            the complete file path
        """
        filenames = Gtk.FileChooserDialog.get_filenames(self)
        for filename in filenames:
            if not path.exists(filename):
                self.show_missing_message(filename)
                return None  # rerun
        return filenames


class OpenFlowGraph(OpenFileDialog):
    title = 'Open a Flow Graph from a File...'
    filter_label = 'Flow Graph Files'
    filter_ext = Constants.FILE_EXTENSION

    def __init__(self, parent, current_file_path=''):
        super(OpenFlowGraph, self).__init__(parent, current_file_path)
        self.set_select_multiple(True)


class OpenQSS(OpenFileDialog):
    title = 'Open a QSS theme...'
    filter_label = 'QSS Themes'
    filter_ext = '.qss'


class SaveFlowGraph(SaveFileDialog):
    title = 'Save a Flow Graph to a File...'
    filter_label = 'Flow Graph Files'
    filter_ext = Constants.FILE_EXTENSION


class SaveConsole(SaveFileDialog):
    title = 'Save Console to a File...'
    filter_label = 'Test Files'
    filter_ext = '.txt'


class SaveScreenShot(SaveFileDialog):
    title = 'Save a Flow Graph Screen Shot...'
    filters = [('PDF Files', '.pdf'), ('PNG Files', '.png'), ('SVG Files', '.svg')]
    filter_ext = '.pdf'  # the default

    def __init__(self, parent, current_file_path=''):
        super(SaveScreenShot, self).__init__(parent, current_file_path)

        self.config = Gtk.Application.get_default().config

        self._button = button = Gtk.CheckButton(label='Background transparent')
        self._button.set_active(self.config.screen_shot_background_transparent())
        self.set_extra_widget(button)

    def setup_filters(self, filters=None):
        super(SaveScreenShot, self).setup_filters(self.filters)

    def show_missing_message(self, filename):
        Dialogs.MessageDialogWrapper(
            self.parent,
            Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE, 'Can not Save!',
            'File Extension of <b>{filename}</b> not supported!'.format(filename=Utils.encode(filename)),
        ).run_and_destroy()

    def run(self):
        valid_exts = {ext for label, ext in self.filters}
        filename = None
        while True:
            response = Gtk.FileChooserDialog.run(self)
            if response != Gtk.ResponseType.OK:
                filename = None
                break

            filename = self.get_filename()
            if path.splitext(filename)[1] in valid_exts:
                break

            self.show_missing_message(filename)

        bg_transparent = self._button.get_active()
        self.config.screen_shot_background_transparent(bg_transparent)
        self.destroy()
        return filename, bg_transparent



# Copyright 2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import

import six


class Flags(object):

    THROTTLE = 'throttle'
    DISABLE_BYPASS = 'disable_bypass'
    NEED_QT_GUI = 'need_qt_gui'
    DEPRECATED = 'deprecated'
    NOT_DSP = 'not_dsp'
    SHOW_ID = 'show_id'
    HAS_PYTHON = 'python'
    HAS_CPP = 'cpp'

    def __init__(self, flags=None):
        if flags is None:
            flags = set()
        if isinstance(flags, six.string_types):
            flags = (f.strip() for f in flags.replace(',', '').split())
        self.data = set(flags)

    def __getattr__(self, item):
        return item in self

    def __contains__(self, item):
        return item in self.data

    def __str__(self):
        return ', '.join(self.data)

    def set(self, *flags):
        self.data.update(flags)



# from https://hg.python.org/cpython/file/default/Lib/collections/__init__.py

from collections import MutableMapping


class ChainMap(MutableMapping):
    """ A ChainMap groups multiple dicts (or other mappings) together
    to create a single, updateable view.

    The underlying mappings are stored in a list.  That list is public and can
    be accessed or updated using the *maps* attribute.  There is no other
    state.

    Lookups search the underlying mappings successively until a key is found.
    In contrast, writes, updates, and deletions only operate on the first
    mapping.

    """

    def __init__(self, *maps):
        """Initialize a ChainMap by setting *maps* to the given mappings.
        If no mappings are provided, a single empty dictionary is used.

        """
        self.maps = list(maps) or [{}]          # always at least one map

    def __missing__(self, key):
        raise KeyError(key)

    def __getitem__(self, key):
        for mapping in self.maps:
            try:
                return mapping[key]             # can't use 'key in mapping' with defaultdict
            except KeyError:
                pass
        return self.__missing__(key)            # support subclasses that define __missing__

    def get(self, key, default=None):
        return self[key] if key in self else default

    def __len__(self):
        return len(set().union(*self.maps))     # reuses stored hash values if possible

    def __iter__(self):
        return iter(set().union(*self.maps))

    def __contains__(self, key):
        return any(key in m for m in self.maps)

    def __bool__(self):
        return any(self.maps)

    def __repr__(self):
        return '{0.__class__.__name__}({1})'.format(
            self, ', '.join(map(repr, self.maps)))

    @classmethod
    def fromkeys(cls, iterable, *args):
        """Create a ChainMap with a single dict created from the iterable."""
        return cls(dict.fromkeys(iterable, *args))

    def copy(self):
        """New ChainMap or subclass with a new copy of maps[0] and refs to maps[1:]"""
        return self.__class__(self.maps[0].copy(), *self.maps[1:])

    __copy__ = copy

    def new_child(self, m=None):                # like Django's Context.push()
        """New ChainMap with a new map followed by all previous maps.
        If no map is provided, an empty dict is used.
        """
        if m is None:
            m = {}
        return self.__class__(m, *self.maps)

    @property
    def parents(self):                          # like Django's Context.pop()
        """New ChainMap from maps[1:]."""
        return self.__class__(*self.maps[1:])

    def __setitem__(self, key, value):
        self.maps[0][key] = value

    def __delitem__(self, key):
        try:
            del self.maps[0][key]
        except KeyError:
            raise KeyError('Key not found in the first mapping: {!r}'.format(key))

    def popitem(self):
        """Remove and return an item pair from maps[0]. Raise KeyError is maps[0] is empty."""
        try:
            return self.maps[0].popitem()
        except KeyError:
            raise KeyError('No keys found in the first mapping.')

    def pop(self, key, *args):
        """Remove *key* from maps[0] and return its value. Raise KeyError if *key* not in maps[0]."""
        try:
            return self.maps[0].pop(key, *args)
        except KeyError:
            raise KeyError('Key not found in the first mapping: {!r}'.format(key))

    def clear(self):
        """Clear maps[0], leaving maps[1:] intact."""
        self.maps[0].clear()



# Copyright 2008-2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


from __future__ import absolute_import

import os

from mako.template import Template

from .hier_block import HierBlockGenerator, QtHierBlockGenerator
from .top_block import TopBlockGenerator
from .cpp_top_block import CppTopBlockGenerator
from .cpp_hier_block import CppHierBlockGenerator

DATA_DIR = os.path.dirname(__file__)
FLOW_GRAPH_TEMPLATE = os.path.join(DATA_DIR, 'flow_graph.py.mako')
flow_graph_template = Template(filename=FLOW_GRAPH_TEMPLATE)


class Generator(object):
    """Adaptor for various generators (uses generate_options)"""

    def __init__(self, flow_graph, file_path):
        """
        Initialize the generator object.
        Determine the file to generate.

        Args:
            flow_graph: the flow graph object
            file_path: the path to the grc file
        """
        self.generate_options = flow_graph.get_option('generate_options')
        self.output_language = flow_graph.get_option('output_language')

        if self.output_language == 'python':

            if self.generate_options == 'hb':
                generator_cls = HierBlockGenerator
            elif self.generate_options == 'hb_qt_gui':
                generator_cls = QtHierBlockGenerator
            else:
                generator_cls = TopBlockGenerator

        elif self.output_language == 'cpp':

            if self.generate_options == 'hb':
                generator_cls = CppHierBlockGenerator
            elif self.generate_options == 'hb_qt_gui':
                pass
            else:
                generator_cls = CppTopBlockGenerator

        self._generator = generator_cls(flow_graph, file_path)

    def __getattr__(self, item):
        """get all other attrib from actual generator object"""
        return getattr(self._generator, item)



"""
Copyright 2008, 2009, 2011 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import

import os
import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GObject

from .Constants import DEFAULT_CONSOLE_WINDOW_WIDTH
from .Dialogs import TextDisplay, MessageDialogWrapper

from ..core import Messages


log = logging.getLogger(__name__)


class Console(Gtk.ScrolledWindow):
    def __init__(self):
        Gtk.ScrolledWindow.__init__(self)
        log.debug("console()")
        self.app = Gtk.Application.get_default()

        self.text_display = TextDisplay()

        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.add(self.text_display)
        self.set_size_request(-1, DEFAULT_CONSOLE_WINDOW_WIDTH)

    def add_line(self, line):
        """
        Place line at the end of the text buffer, then scroll its window all the way down.

        Args:
            line: the new text
        """
        self.text_display.insert(line)
